/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <dlfcn.h>

#include <string>
#include <fstream>
#include <vector>

#include "utils.h"
#include "log.h"
#include "launcher.h"
#include "dotnet_launcher.h"

namespace tizen {
namespace runtime {
namespace dotnetcore {

CoreRuntime::CoreRuntime() :
	initializeClr(nullptr),
	executeAssembly(nullptr),
	shutdown(nullptr),
	createDelegate(nullptr),
	__coreclrLib(nullptr),
	__hostHandle(nullptr),
	__domainId(-1),
	preparedFunction(nullptr),
	launchFunction(nullptr)
{
#define __XSTR(x) #x
#define __STR(x) __XSTR(x)

#ifdef DEVICE_API_DIR
	__deviceAPIDirectory = __STR(DEVICE_API_DIR);
#endif
#ifdef RUNTIME_DIR
	__runtimeDirectory = __STR(RUNTIME_DIR);
#endif
#ifdef NATIVE_LIB_DIR
	__nativeLibDirectory = __STR(NATIVE_LIB_DIR);
#endif

#ifdef USE_MANAGED_LAUNCHER
#ifdef CORECLR_LAUNCHER_ASSEMBLY_PATH
	__launcherAssembly = __STR(CORECLR_LAUNCHER_ASSEMBLY_PATH);
#endif
#endif

#undef __STR
#undef __XSTR

	_DBG("Constructor called!!");
}

CoreRuntime::~CoreRuntime()
{
	dispose();
}

int CoreRuntime::initialize(bool standalone)
{
#ifdef __arm__
	// libunwind library is used to unwind stack frame, but libunwind for ARM
	// does not support ARM vfpv3/NEON registers in DWARF format correctly.
	// Therefore let's disable stack unwinding using DWARF information
	// See https://github.com/dotnet/coreclr/issues/6698
	//
	// libunwind use following methods to unwind stack frame.
	// UNW_ARM_METHOD_ALL          0xFF
	// UNW_ARM_METHOD_DWARF        0x01
	// UNW_ARM_METHOD_FRAME        0x02
	// UNW_ARM_METHOD_EXIDX        0x04
	putenv(const_cast<char *>("UNW_ARM_UNWIND_METHOD=6"));
#endif // __arm__

	if (standalone) {
		const char *deviceApiDirectory = getenv("__deviceAPIDirectory");
		const char *runtimeDirectory = getenv("__runtimeDirectory");
		if (deviceApiDirectory != nullptr)
			__deviceAPIDirectory = deviceApiDirectory;
		if (runtimeDirectory != nullptr)
			__runtimeDirectory = runtimeDirectory;

#ifdef USE_MANAGED_LAUNCHER
		const char *launcherAssembly = getenv("__launcherAssembly");
		if (launcherAssembly != nullptr)
			__launcherAssembly = launcherAssembly;
#endif
	}

	if (__deviceAPIDirectory.empty()) {
		_ERR("Empty Device API Directory");
		return 1;
	} else {
		__deviceAPIDirectory = absolutePath(__deviceAPIDirectory);
	}

	if (__runtimeDirectory.empty()) {
		_ERR("Empty Runtime Directory");
		return 1;
	} else {
		__runtimeDirectory = absolutePath(__runtimeDirectory);
	}

#ifdef USE_MANAGED_LAUNCHER
	if (__launcherAssembly.empty()) {
		_ERR("Empty Launcher Assembly");
		return 1;
	} else {
		__launcherAssembly = absolutePath(__launcherAssembly);
	}
#endif

	std::string libCoreclr(concatPath(__runtimeDirectory, "libcoreclr.so"));

	_DBG("libcoreclr : %s", libCoreclr.c_str());

	__coreclrLib = dlopen(libCoreclr.c_str(), RTLD_NOW | RTLD_LOCAL);
	if (__coreclrLib == nullptr) {
		char *err = dlerror();
		_ERR("dlopen failed to open libcoreclr.so with error %s", err);
		return 1;
	}

#define CORELIB_RETURN_IF_NOSYM(type, variable, name) \
	do { \
		variable = (type)dlsym(__coreclrLib, name); \
		if (variable == nullptr) { \
			_ERR(name " is not found in the libcoreclr.so"); \
			return 1; \
		} \
	} while (0)

	CORELIB_RETURN_IF_NOSYM(coreclr_initialize_ptr, initializeClr, "coreclr_initialize");
	CORELIB_RETURN_IF_NOSYM(coreclr_execute_assembly_ptr, executeAssembly, "coreclr_execute_assembly");
	CORELIB_RETURN_IF_NOSYM(coreclr_shutdown_ptr, shutdown, "coreclr_shutdown");
	CORELIB_RETURN_IF_NOSYM(coreclr_create_delegate_ptr, createDelegate, "coreclr_create_delegate");

#undef CORELIB_RETURN_IF_NOSYM

	_DBG("libcoreclr dlopen and dlsym success");
	_DBG("this addr : %x", this);
	_DBG("coreclr_initialize : %x", initializeClr);

	return 0;
}

bool CoreRuntime::initializeCoreClr(const char* appId,
									 const char* assemblyProbePaths,
									 const char* pinvokeProbePaths,
									 const char* tpaList)
{
	const char *propertyKeys[] = {
		"TRUSTED_PLATFORM_ASSEMBLIES",
		"APP_PATHS",
		"APP_NI_PATHS",
		"NATIVE_DLL_SEARCH_DIRECTORIES",
		"AppDomainCompatSwitch"
	};

	const char *propertyValues[] = {
		tpaList,
		assemblyProbePaths,
		assemblyProbePaths,
		pinvokeProbePaths,
		"UseLatestBehaviorWhenTFMNotSpecified"
	};

	std::string selfPath = readSelfPath();

	int st = initializeClr(selfPath.c_str(),
							appId,
							sizeof(propertyKeys) / sizeof(propertyKeys[0]),
							propertyKeys,
							propertyValues,
							&__hostHandle,
							&__domainId);

	if (st < 0) {
		_ERR("initialize core clr fail! (0x%08x)", st);
		return false;
	}

	_DBG("Initialize core clr success");
	return true;
}

int CoreRuntime::runManagedLauncher(const char* appId, const char* appBase, const char* tpaList)
{
	if (fileNotExist(__launcherAssembly)) {
		_ERR("Launcher assembly is not exist in %s", __launcherAssembly.c_str());
		return 1;
	}

	if (!initializeCoreClr(appId, appBase, appBase, tpaList)) {
		_ERR("Failed to initialize coreclr");
		return 1;
	}

#ifdef USE_MANAGED_LAUNCHER
	void *preparedFunctionDelegate;
	int st = createDelegate(__hostHandle, __domainId,
							"Tizen.Runtime",
							"Tizen.Runtime.Coreclr.AssemblyManager",
							"Prepared", &preparedFunctionDelegate);
	if (st < 0) {
		_ERR("Create delegate for Launch prepared function is fail (0x%08x)", st);
		return 1;
	}
	preparedFunction = reinterpret_cast<PreparedFunctionPtr>(preparedFunctionDelegate);

	if (preparedFunction != nullptr)
		preparedFunction();

	void *launchFunctionDelegate;
	st = createDelegate(__hostHandle, __domainId,
						"Tizen.Runtime",
						"Tizen.Runtime.Coreclr.AssemblyManager",
						"Launch", &launchFunctionDelegate);
	if (st < 0) {
		_ERR("Create delegate for Launch managed function is fail! (0x%08x)", st);
		return 1;
	}
	launchFunction = reinterpret_cast<LaunchFunctionPtr>(launchFunctionDelegate);
#endif
	return 0;
}

void CoreRuntime::dispose()
{
	if (__hostHandle != nullptr) {
		int st = shutdown(__hostHandle, __domainId);
		if (st < 0)
			_ERR("shutdown core clr fail! (0x%08x)", st);
	}

	if (dlclose(__coreclrLib) != 0)
		_ERR("libcoreclr.so close failed");

	__coreclrLib = nullptr;

	_DBG("Dotnet runtime disposed");
}

int CoreRuntime::launch(const char* appId, const char* root, const char* path, int argc, char* argv[])
{
	if (path == nullptr) {
		_ERR("executable path is null");
		return 1;
	}

	if (fileNotExist(path)) {
		_ERR("File not exist : %s", path);
		return 1;
	}

	std::string tpa;
	std::string appRoot = root;
	std::string appBin = concatPath(appRoot, "bin");
	std::string appLib = concatPath(appRoot, "lib");
	std::string probePath = appBin + ":" + appLib + ":" + __nativeLibDirectory;

	std::vector<std::string> searchDirectories;
	searchDirectories.push_back(appBin);
	searchDirectories.push_back(appLib);
	searchDirectories.push_back(__runtimeDirectory);
	searchDirectories.push_back(__deviceAPIDirectory);
#ifdef USE_MANAGED_LAUNCHER
	searchDirectories.push_back(baseName(__launcherAssembly));
#endif

	assembliesInDirectory(searchDirectories, tpa);

#ifdef USE_MANAGED_LAUNCHER
	runManagedLauncher(appId, probePath.c_str(), tpa.c_str());

	bool success = false;
	if (launchFunction != nullptr) {
		std::string cppPath(path);

		if (isManagedAssembly(cppPath) && !isNativeImage(cppPath)) {
			size_t extindex = cppPath.size() - 4;
			cppPath = cppPath.substr(0, extindex) + ".ni" + cppPath.substr(extindex, 4);
			if (!fileNotExist(cppPath))
				path = cppPath.c_str();
		}

		success = launchFunction(root, path, argc, argv);
		if (!success)
			_ERR("Failed to launch Application %s", path);
		return success ? 0 : 1;
	} else {
		_ERR("Failed to find launch function");
		return 1;
	}
#else
	int st = initializeCoreClr(appId, probePath.c_str(), probePath.c_str(), tpa.c_str());
	unsigned int ret = 0;
	st = executeAssembly(__hostHandle, __domainId, argc, (const char**)argv, path, &ret);
	if (st < 0)
		_ERR("Failed to Execute Assembly %s (0x%08x)", path, st);
	return ret;
#endif
}

}  // namespace dotnetcore
}  // namespace runtime
}  // namespace tizen
